/*  File: vec2ps.c  */

#ifdef __MSDOS__
#include <dir.h>
#endif

#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <values.h>

#define BUFSIZE 1024

char
    buffer [BUFSIZE + 1],
    label [BUFSIZE + 1],
    file [BUFSIZE + 1],
    *infile = NULL,
    *programname;

int
    lowdef = 0,
    highdef = 0,
    circle = 0;

float
    low,
    high;

FILE
    *fp;

void
    get_programname (char const *argv0),
    syntax (void),
    process_args (int argc, char *argv []),
    process (void),
    to_show (char *dest, char const *src),
    errit (char const *format, ...);
char
    *get_arg (int *argc, char *argv [], int *index),
    *getline (void);

int main (int argc, char *argv [])
{
    get_programname (argv [0]);

    process_args (argc, argv);

    if (! infile)
        errit ("Naam input-file ontbreekt");

    fp = fopen (infile, "r");
    if (! fp)
        errit ("File \"%s\" not found", infile);

    to_show (file, infile);

    process ();

    fclose (fp);

    return 0;
}

void process ()
{
    int
        size,
        nr_vec,
        s_len,
        i;
    float
        min,
        max,
        f,
        value;

    getline ();
    size = atoi (buffer);
    if (size < 1 || size > 100)
        errit ("Illegal vectorsize (%s)", buffer);

    min = MAXFLOAT,
    max = -MAXFLOAT;
    s_len = nr_vec = 0;
    while (getline ()) {
        nr_vec++;
        if ((i = strlen (buffer)) > s_len)
            s_len = i;
        for (i = 0; i < size; i++) {
            getline ();
            sscanf (buffer, " %f", &value);
            if (value > max)
                max = value;
            if (value < min)
                min = value;
        }
    }
    if (min == max) {
        min -= 0.5;
        max += 0.5;
    }
    if (! lowdef)
        low = ((min < 0.0) ? min : 0.0);
    if (! highdef)
        high = ((max > 1.0) ? max : 1.0);
    if (low > high) {
        low = ((min < 0.0) ? min : 0.0);
        high = ((max > 1.0) ? max : 1.0);
    }

    if (min < low)
        fprintf (stderr, "Scaling: [%f, %f] -> 0\n", min, low);
    fprintf (stderr, "Scaling: [%f, %f] -> [0, 1]\n", low, high);
    if (high < max)
        fprintf (stderr, "Scaling: [%f, %f] -> 1\n", high, max);

    rewind (fp);
    getline ();

    printf (
        "%c!PS\n"
        "%c%cBoundingBox: 72 %i %i 714\n"
        "%c%cCreator: %s, (c) P. Kleiweg 1996\n"
        "%c%cTitle: %s\n"
        "%c%cEndComments\n"
        "\n"
        "/Title (%s) def\n"
        "\n"
        "/Width 14 def\n"
        "/Height 14 def\n"
        "/Margin 72 def\n"
        "\n",
        '%', '%', '%',
        700 - 14 * nr_vec,
        78 + 14 * size + 5 * s_len,
        '%', '%', programname,
        '%', '%', infile,
        '%', '%',
        file
    );
    printf (
        "0.5 setlinewidth\n"
        "\n"
        "/Times-Roman findfont 10 scalefont setfont\n"
        "\n"
        "/v {\n"
    );

    if (circle)
        printf (
            "  sqrt\n"
            "  gsave\n"
            "    Width 2 div Height 2 div rmoveto\n"
            "    Width Height lt { Width } { Height } ifelse 2 div mul /r exch def\n"
            "    currentpoint\n"
            "    r 0 rmoveto\n"
            "    r 0 360 arc fill\n"
            "  grestore\n"
            "  Width 0 rmoveto\n"
        );
    else
        printf (
            "  neg 1.0 add setgray\n"
            "  currentpoint\n"
            "  Width 0 rlineto\n"
            "  0 Height rlineto\n"
            "  Width neg 0 rlineto\n"
            "  closepath\n"
            "  fill\n"
            "  moveto\n"
            "  currentpoint\n"
            "  0 setgray\n"
            "  Width 0 rlineto\n"
            "  0 Height rlineto\n"
            "  Width neg 0 rlineto\n"
            "  closepath\n"
            "  stroke\n"
            "  moveto\n"
            "  Width 0 rmoveto\n"
        );

    printf (
        "} bind def\n"
        "\n"
        "/l {\n"
        "  gsave\n"
        "    ml SHIFT rmoveto\n"
        "    show\n"
        "  grestore\n"
        "  Margin\n"
        "  currentpoint exch pop\n"
        "  Height sub moveto\n"
        "} bind def\n"
        "\n"
        "/ml (w) stringwidth pop def\n"
        "\n"
        "Margin 700 moveto\n"
        "\n"
    );

    printf (
        "/SHIFT Height 2 div\n"
        "gsave\n"
        "  newpath\n"
        "  0 0 moveto\n"
        "  (-) false charpath\n"
        "  pathbbox\n"
        "grestore\n"
        "exch pop add exch pop\n"
        "2 div\n"
        "sub def\n"
        "\n"
    );

    while (getline ()) {
        to_show (label, buffer);
        for (i = 0; i < size; i++) {
            f = (atof (getline ()) - low) / (high - low);
            if (f < 0.0)
                f = 0.0;
            else if (f > 1.0)
                f = 1.0;
            printf ("%f v\n", f);
	}
        printf ("(%s) l\n\n", label);
    }

    printf (
        "0 SHIFT rmoveto\n"
        "Title show\n\n"
        "showpage\n"
    );
}

char *getline ()
{
    int
        i;

    do {
        if (! fgets (buffer, BUFSIZE, fp))
            return NULL;
        i = strlen (buffer);
        while (i && isspace (buffer [i - 1]))
            buffer [--i] = '\0';
        i = 0;
        while (buffer [i] && isspace (buffer [i]))
            i++;
        memmove (buffer, buffer + i, strlen (buffer) + 1);
	if (buffer [0] == '#')
	    buffer [0] = '\0';
    } while (strlen (buffer) == 0);

    return buffer;
}

void to_show (char *dest, char const *src)
/* dest != src
   copy src to dest
   ( -> \(
   ) -> \)
   \ -> \\
*/
{
    int
        i, j;

    i = j = -1;
    while (src [++i])
        switch (src [i]) {
            case '(':
            case ')':
            case '\\':
                dest [++j] = '\\';
                /*  no break  */
    	default:
                dest [++j] = src [i];
	}
    dest [++j] = '\0';
}

void syntax ()
{
    fprintf (
        stderr,
        "\n(c) P. Kleiweg 1996\n"
        "\nSyntax: %s -i file.vec [-c] [-l float] [-u float]\n"
        "  -c: area of circle\n"
        "  -l: lower bound\n"
        "  -u: upper bound\n",
        programname
    );
    exit (1);
}

void process_args (int argc, char *argv [])
{
    int
	index;

    if (argc == 1)
        syntax ();

    for (index = 1; index < argc; index++) {
	if (argv [index][0] == '-') {
	    switch (argv [index][1]) {
		case 'i':
		    infile = strdup (get_arg (&argc, argv , &index));
		    break;
		case 'c':
		    circle = 1;
		    break;
		case 'l':
		    low = atof (get_arg (&argc, argv , &index));
                    lowdef = 1;
		    break;
		case 'u':
		    high = atof (get_arg (&argc, argv , &index));
                    highdef = 1;
		    break;
		default:
                    errit ("Invalid option '%s'", argv [index]);
	    }
	} else
            errit ("Invalid parameter \"%s\"", argv [index]);
    }
}

char *get_arg (int *argc, char *argv [], int *index)
{
    int
	i;
    i = 1;
    while (argv [*index][++i]) {
	if (! isspace (argv [*index][i]))
	    return argv [*index] + i;
    }
    if (*index == *argc - 1)
	errit ("Missing argument for \"%s\"", argv [*index]);
    return argv [++*index];
}

void get_programname (char const *argv0)
{
#ifdef __MSDOS__
    char
        name [MAXFILE];
    fnsplit (argv0, NULL, NULL, name, NULL);
    programname = strdup (name);
#else   /* unix */
    char
        *p;
    p = strrchr (argv0, '/');
    if (p)
        programname = strdup (p + 1);
    else
        programname = strdup (argv0);
#endif    
}

void errit (char const *format, ...)
{
    va_list
	list;

    fprintf (stderr, "\nError %s: ", programname);

    va_start (list, format);
    vfprintf (stderr, format, list);

    fprintf (stderr, "\n\n");

    exit (1);
}
